//////////////////////////////////////////////////////////////////////////////////
//							Abacus(SOROBAN) Program								//
// - Platform	: TI89 or TI91			- Author	: Nobuhiro Kuroiwa			//
// - Compiler	: TIGCC					- Created	: 11/23/01; 20:57:41		//
// - Language	: C						- Contact	: nobu@iastate.edu			//
// - Tabsize    : 4						- Version	: a2-g2						//
// - What haven't been done														//
//				: Check comments												//
//				: Compression of BMP data										//
//////////////////////////////////////////////////////////////////////////////////

// #define OPTIMIZE_ROM_CALLS // Use ROM Call Optimization

#define SAVE_SCREEN           // Save/Restore LCD Contents

#include <tigcclib.h>         // Include All Header Files
#include "abacus.h"

short _ti89;                  // Produce .89Z File
short _ti92plus;              // Produce .9XZ File

//////////////////////////////////////////////////////////////////////////////////
//							Main Function										//
//////////////////////////////////////////////////////////////////////////////////
void _main(void)
{
	TYPE_KEY	key;		// stores key
	BOOL		mode = MODE_PLUS;// default mode is plus
	//	Welcome Message
	ClrScr();				// clear the screen
	DrawStr( 30	, 20, "Abucus ver a2-g2",			A_NORMAL);
	DrawStr( 0	, 40, "press any key to continue",	A_NORMAL);
	DrawStr( 0	, 50, "otherwise press ESC key",	A_NORMAL);
	DrawStr( 0	, 60, "ARROW keys : move cursor",	A_NORMAL);
	DrawStr( 0	, 70, "ENTER key  : move piece",	A_NORMAL);
	DrawStr( 0	, 80, "CLEAR key  : reset pieces",	A_NORMAL);
	DrawStr( 0	, 90, "ESC        : exit",			A_NORMAL);
	
	// while user looking welcome message, do this
	if( !initialize())				return;
	//	ESC to exit
	if( getUserInput() == KEY_ESC)
	{
		GrayOff();
		return;	// ESC to exit
	}
	ClrScr();
	// Otherwise, Let's dance!
	drawScreen();							// put everything on a screen
	// gazing into messed columns
	OSRegisterTimer( USER_TIMER, 20);		// set the timer 1 sec, 1 sec = 20
	waitTillTimerExpired();
	// clean up! this is necessary to call before game roop,
	// cause focus is initialized in this function
	clearPiece();
	// game roop, roop till esc key is pressed
	while( KEY_ESC != ( key = getUserInput()))
	{
		if		( key == KEY_ENTER)			movePiece();
		else if	( key == KEY_CLEAR)			clearPiece();
		else if	( 49 <= key && key <= 57)	// if the key is number key
		{
			if( mode == MODE_MINUS)	// subtraction
				sub( -1*(key - 48), focus.pcol_current);
			else					// addition
				add( key - 48, focus.pcol_current);
		}
		else if	( key == '+')
		{
			drawColumn( focus.pcol_current, SPRT_XOR);	// erase focused piece
			focus.bmp = &bmp_plus;						// update focuse
			drawColumn( focus.pcol_current, SPRT_OR);	// draw new focus
			mode = MODE_PLUS;							// set mode to plusmode
		}
		else if	( key == '-')
		{
			drawColumn( focus.pcol_current, SPRT_XOR);	// erase focused piece
			focus.bmp = &bmp_minus;						// update focus
			drawColumn( focus.pcol_current, SPRT_OR);	// draw new focus
			mode = MODE_MINUS;							// set mode to minusmode
		}				
		else	moveFocus( key);						// move focus	
	}
	GrayOff ();
}

//////////////////////////////////////////////////////////////////////////////////
//							waitTillTimerExpired()								//
//////////////////////////////////////////////////////////////////////////////////
void waitTillTimerExpired()
{
	while ( !OSTimerExpired( USER_TIMER));
}


//////////////////////////////////////////////////////////////////////////////////
//							getUserInput()										//
// get input key pressed by user											 	//
//																		 	  	//	
//	replacement for ngetchx() and/or GKeyIn(), because:					 	  	//
//																		 	  	//
// (1) ngetchx() won't check the APD timer value							 	//
// (2) GKeyIn() destroys the display due to setting the status display		   	//
//////////////////////////////////////////////////////////////////////////////////
int getUserInput() {
	void *kbq = kbd_queue();
	int key;

	OSTimerRestart(APD_TIMER);
	while( OSdequeue( &key, kbq))
	{
		if (OSTimerExpired(APD_TIMER)) {
			off();
			OSTimerRestart(APD_TIMER);
		}
	}
	if( key == KEY_OFF)
	{
		off();
		OSTimerRestart(APD_TIMER);
		return 32;				// just avoid caution
	}
	else {
		return key & 0xf7ff; 	// mask out 0x800 (repeat flag)
	}
}

//////////////////////////////////////////////////////////////////////////////////
//							grayClrScr()										//
//	Clear the screen															//
//////////////////////////////////////////////////////////////////////////////////
void grayClrScr()
{
	// clear light plane
	SetPlane( LIGHT_PLANE);
	ScrRectFill (&(SCR_RECT){{0, 0, 160, 100}}, &(SCR_RECT){{0, 0, 159, 99}}, A_REVERSE);
	// clear dark plane
	SetPlane ( DARK_PLANE);
	ScrRectFill (&(SCR_RECT){{0, 0, 160, 100}}, &(SCR_RECT){{0, 0, 159, 99}}, A_REVERSE); 
}

//////////////////////////////////////////////////////////////////////////////////
//							clearPiece()										//
//////////////////////////////////////////////////////////////////////////////////
void clearPiece()
{
	int i;
	// set focus to most left up
	focus.pcol_current = &col[0];	focus.pos_focus = 0;
	// zero clear the columns
	for( i = 0; i < BLOCKxCOLUMN; i++)	setNumber( &col[i], ( i + 1) % 2);
	// redraw screen. drawScreen() includes clearing screen
	drawScreen();
	OSFreeTimer( USER_TIMER);// ready the timer
	OSRegisterTimer( USER_TIMER, 5);// set timer to 100ms
	// move pieces in up column every 100ms
	for( i = 0; i < NUM_COLUMN; i++)
	{
		OSTimerRestart( USER_TIMER);
		waitTillTimerExpired();
		movePiece();
		moveFocus( KEY_RIGHT);
	}
	// set focus to default position
	jumpFocus( &col[9], 0);
	// release timer
	OSFreeTimer( USER_TIMER);
}

//////////////////////////////////////////////////////////////////////////////////
//								drawScreen()									//
//////////////////////////////////////////////////////////////////////////////////
void drawScreen()
{
	int i;
	grayClrScr();
	// draw every column
	for( i = 0; i < BLOCKxCOLUMN; i++)		drawColumn( &col[i], SPRT_OR);
	SetPlane( LIGHT_PLANE);
	//	draw an Insulation separates Up column and Down column
	DrawLine(
		0, 		PIECE_HEIGHT + AXIS_HEIGHT,
		159 , 	PIECE_HEIGHT + AXIS_HEIGHT,
		A_NORMAL
	);
	DrawLine(
		0, 		PIECE_HEIGHT + AXIS_HEIGHT + INSUL_HEIGHT - 1,
		159 , 	PIECE_HEIGHT + AXIS_HEIGHT + INSUL_HEIGHT - 1,
		A_NORMAL
	);
	SetPlane( DARK_PLANE);
	DrawLine(
		0, 		PIECE_HEIGHT + AXIS_HEIGHT,
		159 , 	PIECE_HEIGHT + AXIS_HEIGHT,
		A_NORMAL
	);
	DrawLine(
		0, 		PIECE_HEIGHT + AXIS_HEIGHT + INSUL_HEIGHT - 1,
		159 , 	PIECE_HEIGHT + AXIS_HEIGHT + INSUL_HEIGHT - 1,
		A_NORMAL
	);
}

//////////////////////////////////////////////////////////////////////////////////
//							initialize()										//
//	initializaton of this program												//
//////////////////////////////////////////////////////////////////////////////////
BOOL initialize()
{
	int i;
	
	//	prepare light and dark plane
	if (!GrayOn ()) return FALSE;	// gray mode on
	lightplane = GetPlane( LIGHT_PLANE);
	darkplane = GetPlane( DARK_PLANE);
	
	// initializa the bitmap objects
	bmp_piece.light		= bmp_data_piece_light;
	bmp_piece.lightgray	= bmp_data_piece_lightgray;
	bmp_piece.gray		= bmp_data_piece_gray;
	bmp_axis.light		= NULL;
	bmp_axis.lightgray	= NULL;
	bmp_axis.gray		= bmp_data_axis;
	bmp_plus.light		= bmp_data_plus_light;
	bmp_plus.lightgray	= bmp_data_plus_lightgray;
	bmp_plus.gray		= bmp_data_plus_gray;
	bmp_minus.light		= bmp_data_minus_light;
	bmp_minus.lightgray	= bmp_data_minus_lightgray;
	bmp_minus.gray		= bmp_data_minus_gray;
	focus.bmp			= &bmp_plus;

	//	shuffle the pieces
	randomize();									// initialize a seed
	for( i = 0; i < NUM_COLUMN; i++)
	{
		col[ i * 2		].pos_axis	= random( 2);	// randomize up column
		col[ i * 2 + 1	].pos_axis	= random( 5);	// randomize down column
	}
	OSFreeTimer( USER_TIMER);						// release timer
	return TRUE;
}

//////////////////////////////////////////////////////////////////////////////////
//							setNumber()											//
//	set a number to target column												//
//	pcol_target	: pointer to target column										//
//	new_axis	: number expected to set to the target column					//
//////////////////////////////////////////////////////////////////////////////////
void setNumber( struct Column *pcol_target, int new_axis)
{
	pcol_target->pos_axis = abs(
								pcol_target->num_piece
								* (int)pcol_target->breverse
								- new_axis
								);
}

//////////////////////////////////////////////////////////////////////////////////
//							getNumber											//
//	get a number from target column												//
//	pcol_target	: pointer of target column										//
//////////////////////////////////////////////////////////////////////////////////
int getNumber( struct Column *pcol_target)
{
	return abs(
				pcol_target->num_piece
				* (int)pcol_target->breverse
				 - pcol_target->pos_axis
			);
}

//////////////////////////////////////////////////////////////////////////////////
//							drawColumn()										//
//	draws a column																//
//	pcol	: pointer to a column wanted to be drawn							//
//	mode	: mode, MODE_PLUS or MODE_MINUS	represents different bmp of focused	//
//			  piece for each mode.												//
//////////////////////////////////////////////////////////////////////////////////
void drawColumn( struct Column *pcol, short mode)
{
	struct Bitmap *pbmp;
	int i, j, dY, cur_pos_y;
	dY = 0;
	cur_pos_y = pcol->y;

	for( i = 0, j = 0; i < pcol->num_piece + 1; i++)
	{
		if( i == pcol->pos_axis)
		{
			pbmp = &bmp_axis;
			dY = AXIS_HEIGHT;
		}
		else
		{
			if( focus.pcol_current == pcol && focus.pos_focus == j++)
				pbmp = focus.bmp;
			else
				pbmp = &bmp_piece;
			dY = PIECE_HEIGHT;
		}
		graySprite32( pcol->x, cur_pos_y, dY, pbmp, mode);
		cur_pos_y += dY;
	}
}

//////////////////////////////////////////////////////////////////////////////////
//							add()												//
//	add number to target column													//
//	num_addition	: number wanted to add. 0 < num_addition < 9				//
//	pcol_target		: target column												//
//////////////////////////////////////////////////////////////////////////////////
BOOL add( int num_addition, struct Column *pcol_target)
{
	int total;
	if( pcol_target == NULL)	return FALSE;
	// get the total number should be
	total = num_addition + getNumber( pcol_target);
	// delete the target column
	drawColumn( pcol_target, SPRT_XOR);
	// set the number in current column as much as possible
	setNumber( pcol_target, total % ( 1 + pcol_target->num_piece));
	// draw the renewed target column
	drawColumn( pcol_target, SPRT_OR);
	// if the number is small enough to store in the column
	if( total <= pcol_target->num_piece)		return TRUE;
	// if the number is too big, store the left number to next column
	return add(	total / ( 1 + pcol_target->num_piece), pcol_target->pcol_up_order);
}

//////////////////////////////////////////////////////////////////////////////////
//							sub()												//
//	subtruct a number from target column										//
//	num_subtraction	: number wanted to subtruct. 0 < num_subtraction < 9		//
//	pcol_target		: target column												//
//////////////////////////////////////////////////////////////////////////////////
BOOL sub( signed int num_subtraction, struct Column *pcol_target)
{
	int total, up1, val;
	BOOL flg_one;
	val = getNumber( pcol_target);

	if( pcol_target == NULL)	return FALSE;
		
	// get the total number should be
	total = val + num_subtraction;
	
	up1	= pcol_target->num_piece + 1;
	
	// delete the target column
	drawColumn( pcol_target, SPRT_XOR);
	
	// set the number in current column as much as possible
	setNumber( pcol_target, abs((up1 + total % up1)%up1));
	// draw the renewed target column
	drawColumn( pcol_target, SPRT_OR);
	
	// if the number is small enough to store in the column
	if( 0 <= total)		return TRUE;
	if( total % up1 == 0)	flg_one = 0;
	else					flg_one = 1;
	return sub( total / up1 - flg_one, pcol_target->pcol_up_order);
}

//////////////////////////////////////////////////////////////////////////////////
//							jumpFocus()											//
//	jump a focus to directed position											//
//	pcol_move_to	: target column the focus move to							//
//	new_pos_focus	: position of the focus on the new column					//
//					  0 < new_pos_focus < number of pieces of column move to	//
//////////////////////////////////////////////////////////////////////////////////
void jumpFocus( struct Column *pcol_move_to, int new_pos_focus)
{
	struct Column *pcol_move_from = focus.pcol_current;

	drawColumn( pcol_move_from, SPRT_XOR);			// delete move from w/ focus
	drawColumn( pcol_move_to, SPRT_XOR);			// delete move to w/o focus
	focus.pcol_current = pcol_move_to;				// update the pointer
	if( new_pos_focus > pcol_move_to->num_piece)	// update pos_focus
		focus.pos_focus = pcol_move_to->num_piece - 1;
	else
		focus.pos_focus = new_pos_focus;
	drawColumn( pcol_move_from, SPRT_OR);			// draw move from w/o focus
	drawColumn( pcol_move_to, SPRT_OR);				// draw move to w/ focus
}

//////////////////////////////////////////////////////////////////////////////////
//								moveFocus()										//
//	moving focus by accepting arrow key code									//
//////////////////////////////////////////////////////////////////////////////////
void moveFocus( TYPE_KEY key)
{
	int new_pos_focus;
	// if up or down
	if( key == KEY_UP)
	{
		// if no longer in the column, move to another column
		if( focus.pos_focus == 0)	// pos < 0 's gonna be error if the type is short
		{
			jumpFocus(
						focus.pcol_current->pcol_up,
						focus.pcol_current->pcol_up->num_piece - 1
					);	// set focus full bottom
			return;
		}
		// if still in the column, update position of the focus
		new_pos_focus = focus.pos_focus - 1;
	}
	else if( key == KEY_DOWN)
	{
		// if no longer in the column, move to another column
		if( focus.pos_focus + 1 >= focus.pcol_current->num_piece)
		{
			jumpFocus( focus.pcol_current->pcol_down, 0);	// set focus full up
			return;
		}
		// if still in the column, update the position of the focus
		new_pos_focus = focus.pos_focus + 1;
	}
	else if( key == KEY_RIGHT)
	{
		jumpFocus( focus.pcol_current->pcol_right, focus.pos_focus);
		return;
	}
	else if( key == KEY_LEFT)
	{
		jumpFocus( focus.pcol_current->pcol_left, focus.pos_focus);
		return;
	}
	else
	{
		return;
	}
	// delete move from
	drawColumn( focus.pcol_current, SPRT_XOR);
	focus.pos_focus = new_pos_focus;
	// draw move to
	drawColumn( focus.pcol_current, SPRT_OR);
}

//////////////////////////////////////////////////////////////////////////////////
//							movePiece()											//
//	moving focused piece														//
//////////////////////////////////////////////////////////////////////////////////
void movePiece()
{
	drawColumn( focus.pcol_current, SPRT_XOR);
	focus.pcol_current->pos_axis
		= focus.pos_focus
		+ ( focus.pcol_current->pos_axis <= focus.pos_focus);
	drawColumn( focus.pcol_current, SPRT_OR);
}

//////////////////////////////////////////////////////////////////////////////////
//							graySprite32()										//
//	draw 32 * height sized bray scale bitmap									//
//	x, y	: position to put the bitmap										//
//	height	: height of the bitmap												//
//	bmp		: pointer to the grayscale bitmap structure							//
//	mode	: SPRT_XOR, SPRT_OR, SPRT_AND										//
//			  SPRT_OR to same position and bitmap deletes the bitmap			//
//////////////////////////////////////////////////////////////////////////////////
void graySprite32( int x,int y, int height, struct Bitmap *bmp, short mode)
{
	if( bmp->light 		!= NULL)	Sprite32( x, y, height, bmp->light, lightplane, mode);
	if( bmp->lightgray	!= NULL)	Sprite32( x, y, height, bmp->lightgray, darkplane, mode);
	Sprite32( x, y, height, bmp->gray, lightplane, mode);
	Sprite32( x, y, height, bmp->gray, darkplane, mode);
}
